"""Aiohttp plugin. Includes a path helper that allows you to pass an AbstractRoute.
Takes the method from the route and docstring from the route handler.
::

    from aiohttp import web
    from apispec import APISpec
    from pprint import pprint


    async def hello(request):
        '''Get a greeting endpoint.
        ---
        description: Get a greeting
        responses:
          200:
            description: A greeting to the client
            content:
                text/plain:
                    schema:
                        $ref: '#/definitions/Greeting'
        '''
        return web.Response(text="hello")


    app = web.Application()
    app.add_routes([web.get("/hello", hello)])

    # Add all aiohttp routes to the APISpec
    for route in app.router.routes():
        # Don't include HEAD mehods in OpenAPI spec
        if route.method == "HEAD":
            continue

        spec.path(
            route=route,
        )

    pprint(spec.to_dict()["paths"])
    # {'/hello': {'get': {'description': 'Get a greeting',
    #                    'responses': {'200': {'content': {'text/plain': {'schema': {'$ref': '#/definitions/Greeting'}}},
    #                                          'description': 'A greeting to the '
    #                                                         'client'}}}}}

"""  # noqa: E501

from typing import Any, List, Optional

from aiohttp.web import AbstractRoute
from apispec import BasePlugin, yaml_utils


class AiohttpPlugin(BasePlugin):
    def path_helper(
        self,
        path: Optional[str] = None,
        operations: Optional[dict] = None,
        parameters: Optional[List[dict]] = None,
        *,
        route: Optional[AbstractRoute] = None,
        **kwargs: Any,
    ) -> Optional[str]:
        """Path helper that allows passing a aiohttp AbstractRoute"""
        assert operations is not None
        assert route is not None

        docstring = route.handler.__doc__ or ""
        operations.update(
            {route.method.lower(): yaml_utils.load_yaml_from_docstring(docstring)}
        )
        return route.resource.canonical



"""Flask plugin. Includes a path helper that allows you to pass a view
function to `path`. Inspects URL rules and view docstrings.

Passing a view function::

    from flask import Flask

    app = Flask(__name__)


    @app.route("/gists/<gist_id>")
    def gist_detail(gist_id):
        '''Gist detail view.
        ---
        x-extension: metadata
        get:
            responses:
                200:
                    schema:
                        $ref: '#/definitions/Gist'
        '''
        return "detail for gist {}".format(gist_id)


    with app.test_request_context():
        spec.path(view=gist_detail)
    print(spec.to_dict()["paths"])
    # {'/gists/{gist_id}': {'get': {'responses': {200: {'schema': {'$ref': '#/definitions/Gist'}}}},
    #                  'x-extension': 'metadata'}}

Passing a method view function::

    from flask import Flask
    from flask.views import MethodView

    app = Flask(__name__)


    class GistApi(MethodView):
        '''Gist API.
        ---
        x-extension: metadata
        '''

        def get(self):
            '''Gist view
            ---
            responses:
                200:
                    schema:
                        $ref: '#/definitions/Gist'
            '''
            pass

        def post(self):
            pass


    method_view = GistApi.as_view("gists")
    app.add_url_rule("/gists", view_func=method_view)
    with app.test_request_context():
        spec.path(view=method_view)

    # Alternatively, pass in an app object as a kwarg
    # spec.path(view=method_view, app=app)

    print(spec.to_dict()["paths"])
    # {'/gists': {'get': {'responses': {200: {'schema': {'$ref': '#/definitions/Gist'}}}},
    #             'post': {},
    #             'x-extension': 'metadata'}}


"""  # noqa: E501

import re
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Union

from apispec import BasePlugin, yaml_utils
from apispec.exceptions import APISpecError
from flask import Flask, current_app
from flask.views import MethodView
from werkzeug.routing import Rule

if TYPE_CHECKING:
    from flask.typing import RouteCallable


# from flask-restplus
RE_URL = re.compile(r"<(?:[^:<>]+:)?([^<>]+)>")


class FlaskPlugin(BasePlugin):
    """APISpec plugin for Flask"""

    @staticmethod
    def flaskpath2openapi(path: str) -> str:
        """Convert a Flask URL rule to an OpenAPI-compliant path.

        :param str path: Flask path template.
        """
        return RE_URL.sub(r"{\1}", path)

    @staticmethod
    def _rule_for_view(
        view: Union[Callable[..., Any], "RouteCallable"],
        app: Optional[Flask] = None,
    ) -> Rule:
        if app is None:
            app = current_app

        view_funcs = app.view_functions
        endpoint = None
        for ept, view_func in view_funcs.items():
            if view_func == view:
                endpoint = ept
        if not endpoint:
            raise APISpecError(f"Could not find endpoint for view {view}")

        # WARNING: Assume 1 rule per view function for now
        rule = app.url_map._rules_by_endpoint[endpoint][0]
        return rule

    def path_helper(
        self,
        path: Optional[str] = None,
        operations: Optional[dict] = None,
        parameters: Optional[List[dict]] = None,
        *,
        view: Optional[Union[Callable[..., Any], "RouteCallable"]] = None,
        app: Optional[Flask] = None,
        **kwargs: Any,
    ) -> Optional[str]:
        """Path helper that allows passing a Flask view function."""
        assert view is not None
        assert operations is not None

        rule = self._rule_for_view(view, app=app)
        view_doc = view.__doc__ or ""
        doc_operations = yaml_utils.load_operations_from_docstring(view_doc)
        operations.update(doc_operations)
        if hasattr(view, "view_class") and issubclass(view.view_class, MethodView):  # noqa: E501
            # method attribute is dynamically added, which is supported by mypy
            for method in view.methods:  # type:ignore[union-attr]
                if rule.methods and method in rule.methods:
                    method_name = method.lower()
                    method = getattr(view.view_class, method_name)
                    method_docstring = method.__doc__ or ""
                    operations[method_name] = yaml_utils.load_yaml_from_docstring(  # noqa: E501
                        method_docstring
                    )
        return self.flaskpath2openapi(rule.rule)



"""Tornado plugin. Includes a path helper that allows you to pass an urlspec (path-handler pair)
object to `path`.
::

    from pprint import pprint

    from tornado.web import RequestHandler


    class HelloHandler(RequestHandler):
        def get(self):
            '''Get a greeting endpoint.
            ---
            description: Get a greeting
            responses:
                200:
                    description: A greeting to the client
                    schema:
                        $ref: '#/definitions/Greeting'
            '''
            self.write("hello")


    urlspec = (r"/hello", HelloHandler)
    spec.path(urlspec=urlspec)
    pprint(spec.to_dict()["paths"])
    # {'/hello': {'get': {'description': 'Get a greeting',
    #                     'responses': {200: {'description': 'A greeting to the '
    #                                                     'client',
    #                                         'schema': {'$ref': '#/definitions/Greeting'}}}}}}

"""  # noqa: E501

import inspect
from typing import Any, Callable, Dict, Iterator, List, Optional, Union, cast

from apispec import BasePlugin, yaml_utils
from apispec.exceptions import APISpecError
from tornado.routing import PathMatches
from tornado.web import RequestHandler, URLSpec


class TornadoPlugin(BasePlugin):
    """APISpec plugin for Tornado"""

    @staticmethod
    def _operations_from_methods(
        handler_class: RequestHandler,
    ) -> Iterator[Dict[str, dict]]:
        """Generator of operations described in handler's http methods

        :param handler_class:
        :type handler_class: RequestHandler descendant
        """
        for httpmethod in sorted(yaml_utils.PATH_KEYS):
            method = getattr(handler_class, httpmethod)
            docstring = method.__doc__ or ""
            operation_data = yaml_utils.load_yaml_from_docstring(docstring)
            if operation_data:
                operation = {httpmethod: operation_data}
                yield operation

    @staticmethod
    def tornadopath2openapi(urlspec: URLSpec, method: Callable) -> str:
        """Convert Tornado URLSpec to OpenAPI-compliant path.

        :param urlspec:
        :type urlspec: URLSpec
        :param method: Handler http method
        :type method: function
        """
        matcher = cast(PathMatches, urlspec.matcher)
        regex = matcher.regex
        path_tpl = cast(str, matcher._path)
        if regex.groups:
            if regex.groupindex:
                # urlspec path uses named groups
                sorted_pairs = sorted(
                    ((k, v) for k, v in regex.groupindex.items()),
                    key=lambda kv: kv[1],
                )
                args = [pair[0] for pair in sorted_pairs]
            else:
                args = list(inspect.signature(method).parameters.keys())[1:]

            params = tuple(f"{{{arg}}}" for arg in args)
            path = path_tpl % params
        else:
            path = path_tpl
        if path.count("/") > 1:
            path = path.rstrip("/?*")
        return path

    @staticmethod
    def _extensions_from_handler(handler_class: RequestHandler) -> dict:
        """Returns extensions dict from handler docstring

        :param handler_class:
        :type handler_class: RequestHandler descendant
        """
        docstring = handler_class.__doc__ or ""
        return yaml_utils.load_yaml_from_docstring(docstring)

    def path_helper(
        self,
        path: Optional[str] = None,
        operations: Optional[dict] = None,
        parameters: Optional[List[dict]] = None,
        *,
        urlspec: Optional[Union[URLSpec, tuple]] = None,
        **kwargs: Any,
    ) -> Optional[str]:
        """Path helper that allows passing a Tornado URLSpec or tuple."""
        assert operations is not None
        assert urlspec is not None

        if not isinstance(urlspec, URLSpec):
            urlspec = URLSpec(*urlspec)
        for operation in self._operations_from_methods(urlspec.handler_class):
            operations.update(operation)
        if not operations:
            raise APISpecError(f"Could not find endpoint for urlspec {urlspec}")  # noqa: E501
        params_method = getattr(
            urlspec.handler_class,
            list(operations.keys())[0],
        )
        operations.update(self._extensions_from_handler(urlspec.handler_class))
        return self.tornadopath2openapi(urlspec, params_method)



"""Bottle plugin. Includes a path helper that allows you to pass
a view function to `path`.
::

    from bottle import route, default_app

    app = default_app()


    @route("/gists/<gist_id>")
    def gist_detail(gist_id):
        '''Gist detail view.
        ---
        get:
            responses:
                200:
                    schema:
                        $ref: '#/definitions/Gist'
        '''
        return "detail for gist {}".format(gist_id)


    spec.path(view=gist_detail)
    print(spec.to_dict()["paths"])
    # {'/gists/{gist_id}': {'get': {'responses': {200: {'schema': {'$ref': '#/definitions/Gist'}}}}}}
"""  # noqa: E501

import re
from typing import Any, Callable, List, Optional

from apispec import BasePlugin, yaml_utils
from apispec.exceptions import APISpecError
from bottle import Bottle, Route, default_app

RE_URL = re.compile(r"<([^<>:]+):?[^>]*>")


_default_app = default_app()


class BottlePlugin(BasePlugin):
    """APISpec plugin for Bottle"""

    @staticmethod
    def bottle_path_to_openapi(path: str) -> str:
        return RE_URL.sub(r"{\1}", path)

    @staticmethod
    def _route_for_view(app: Bottle, view: Callable[..., Any]) -> Route:
        endpoint = None
        for route in app.routes:
            if route.callback == view:
                endpoint = route
                break
        if not endpoint:
            raise APISpecError(f"Could not find endpoint for route {view}")
        return endpoint

    def path_helper(
        self,
        path: Optional[str] = None,
        operations: Optional[dict] = None,
        parameters: Optional[List[dict]] = None,
        *,
        view: Optional[Any] = None,
        **kwargs: Any,
    ) -> Optional[str]:
        """Path helper that allows passing a bottle view function."""
        assert operations is not None
        assert view is not None

        docstring = view.__doc__ or ""
        operations.update(yaml_utils.load_operations_from_docstring(docstring))
        app = kwargs.get("app", _default_app)
        route = self._route_for_view(app, view)
        return self.bottle_path_to_openapi(route.rule)
